package com.marvinelsen.willow.domain

import com.github.houbb.segment.bs.SegmentBs
import com.github.houbb.segment.data.phrase.core.data.SegmentPhraseDatas
import com.github.houbb.segment.data.pos.core.data.SegmentPosDatas
import com.github.houbb.segment.support.format.impl.SegmentFormats
import com.github.houbb.segment.support.segment.impl.Segments
import com.github.houbb.segment.support.segment.mode.impl.SegmentModes
import com.github.houbb.segment.support.segment.result.impl.SegmentResultHandlers
import com.github.houbb.segment.support.tagging.pos.tag.impl.SegmentPosTaggings
import kotlinx.serialization.json.Json
import java.sql.Connection
import java.sql.PreparedStatement
import java.sql.ResultSet

class SqliteDictionary(private val connection: Connection) : Dictionary {
    private val whitespaceRegex = """\s+""".toRegex()

    private val segmentBs = SegmentBs.newInstance()
        .segment(Segments.defaults())
        .segmentData(SegmentPhraseDatas.define())
        .segmentMode(SegmentModes.dict())
        .segmentFormat(SegmentFormats.chineseSimple())
        .posTagging(SegmentPosTaggings.simple())
        .posData(SegmentPosDatas.define())

    private val searchSimplifiedPreparedStatement: PreparedStatement by lazy {
        connection.prepareStatement(
            """
            SELECT traditional, simplified, pinyin_with_tone_marks, pinyin_with_tone_numbers, zhuyin, definitions 
            FROM cedict 
            WHERE simplified GLOB ?
            ORDER BY character_count ASC
            """.trimIndent()
        )
    }

    private val searchTraditionalPreparedStatement: PreparedStatement by lazy {
        connection.prepareStatement(
            """
            SELECT traditional, simplified, pinyin_with_tone_marks, pinyin_with_tone_numbers, zhuyin, definitions 
            FROM cedict 
            WHERE traditional GLOB ?
            ORDER BY character_count ASC
            """.trimIndent()
        )
    }

    private val searchPinyinPreparedStatement: PreparedStatement by lazy {
        connection.prepareStatement(
            """
            SELECT traditional, simplified, pinyin_with_tone_marks, pinyin_with_tone_numbers, zhuyin, definitions 
            FROM cedict 
            WHERE searchable_pinyin GLOB ?
            OR searchable_pinyin_with_tone_numbers GLOB ?
            ORDER BY character_count ASC
            """.trimIndent()
        )
    }

    private val searchSegments = """
            WITH cte(id, segment) AS (VALUES ?)
            SELECT cedict.traditional, simplified, pinyin_with_tone_marks, pinyin_with_tone_numbers, zhuyin, definitions  
            FROM cedict INNER JOIN cte
            ON cte.segment = cedict.traditional OR cte.segment = cedict.simplified
            ORDER BY cte.id
    """.trimIndent()

    private val findWordsContaining: PreparedStatement by lazy {
        connection.prepareStatement(
            """
            SELECT traditional, simplified, pinyin_with_tone_marks, pinyin_with_tone_numbers, zhuyin, definitions 
            FROM cedict
            WHERE traditional LIKE ?
            ORDER BY character_count ASC
            """.trimIndent()
        )
    }

    private val findCharacters = """
            SELECT traditional, simplified, pinyin_with_tone_marks, pinyin_with_tone_numbers, zhuyin, definitions 
            FROM cedict
            WHERE traditional IN (?)
    """.trimIndent()

    override fun search(query: String, searchMode: SearchMode) = when (searchMode) {
        SearchMode.PINYIN -> searchPinyin(query)
        SearchMode.SIMPLIFIED -> searchSimplified(query)
        SearchMode.TRADITIONAL -> searchTraditional(query)
        SearchMode.SEGMENTS -> searchSegments(query)
        SearchMode.ENGLISH -> TODO()
    }

    override fun findWordsContaining(entry: DictionaryEntry): List<DictionaryEntry> {
        findWordsContaining.setString(1, "_%${entry.traditional}%")

        val resultSet: ResultSet = findWordsContaining.executeQuery()

        return resultSet.toListOfDictionaryEntries()
    }

    override fun findSentencesContaining(entry: DictionaryEntry): List<DictionaryEntry> {
        return emptyList()
    }

    override fun findCharacters(entry: DictionaryEntry): List<DictionaryEntry> {
        val characterList = entry.traditional
            .split("")
            .filter { it.isNotBlank() }
            .joinToString(",") { "'$it'" }

        val query = findCharacters.replace("?", characterList)

        val resultSet: ResultSet = connection.createStatement().executeQuery(query)

        return resultSet.toListOfDictionaryEntries()
    }

    override fun searchSegments(phrase: String): List<DictionaryEntry> {
        val segments = segmentBs.segment(phrase, SegmentResultHandlers.word())

        val segmentsListString = segments
            .mapIndexed { index, s -> "($index, '$s')" }
            .joinToString(",")

        val query = searchSegments.replace("?", segmentsListString)

        val resultSet: ResultSet = connection.createStatement().executeQuery(query)

        return resultSet.toListOfDictionaryEntries()
    }

    private fun searchSimplified(query: String): List<DictionaryEntry> {
        searchSimplifiedPreparedStatement.setString(1, "$query*")

        val resultSet: ResultSet = searchSimplifiedPreparedStatement.executeQuery()

        return resultSet.toListOfDictionaryEntries()
    }

    private fun searchPinyin(query: String): List<DictionaryEntry> {
        val sanitizedQuery = query.lowercase().replace(whitespaceRegex, "")

        searchPinyinPreparedStatement.setString(1, "$sanitizedQuery*")
        searchPinyinPreparedStatement.setString(2, "$sanitizedQuery*")

        val resultSet: ResultSet = searchPinyinPreparedStatement.executeQuery()

        return resultSet.toListOfDictionaryEntries()
    }

    private fun searchTraditional(query: String): List<DictionaryEntry> {
        searchTraditionalPreparedStatement.setString(1, "$query*")

        val resultSet: ResultSet = searchTraditionalPreparedStatement.executeQuery()

        return resultSet.toListOfDictionaryEntries()
    }
}

@Suppress("MagicNumber")
private fun ResultSet.toDictionaryEntry() = DictionaryEntry(
    traditional = this.getString(1),
    simplified = this.getString(2),
    pinyinWithToneMarks = this.getString(3),
    pinyinWithToneNumbers = this.getString(4),
    zhuyin = this.getString(5),
    definitions = Json.decodeFromString(this.getString(6))
)

private fun ResultSet.toListOfDictionaryEntries() = buildList {
    this@toListOfDictionaryEntries.use {
        while (it.next()) {
            add(it.toDictionaryEntry())
        }
    }
}
